use std::{
    fs::{self, File},
    io::{BufRead, BufReader, Read, Write},
    path::{Path, PathBuf},
};

use crate::error::Error;

pub trait AsParse {
    fn as_parse(&self) -> &dyn Parse;
    fn as_parse_mut(&mut self) -> &mut dyn Parse;
}

impl<P: Parse> AsParse for P {
    fn as_parse(&self) -> &dyn Parse {
        self
    }

    fn as_parse_mut(&mut self) -> &mut dyn Parse {
        self
    }
}

pub trait Parse: AsParse {
    fn parse(&mut self, path: &str, src: &[u8], dst: &mut Vec<u8>) -> Result<(), Error>;
}

pub struct Parser<'a> {
    steps: Vec<&'a mut dyn Parse>,
}

impl<'a> Parser<'a> {
    pub fn new() -> Self {
        Parser { steps: Vec::new() }
    }

    pub fn push<P: AsParse>(&mut self, parser: &'a mut P) {
        self.steps.push(parser.as_parse_mut());
    }

    pub fn append(&mut self, parsers: &mut Vec<&'a mut dyn Parse>) {
        self.steps.append(parsers);
    }
}

impl<'a> Parse for Parser<'a> {
    fn parse(&mut self, path: &str, src: &[u8], dst: &mut Vec<u8>) -> Result<(), Error> {
        let mut buf_1 = Vec::from(src);
        let mut buf_2 = Vec::from(dst.as_slice());

        self.steps
            .iter_mut()
            .try_fold((&mut buf_1, &mut buf_2), |(src, mut dst), p| {
                dst.clear();
                p.parse(path, src.as_slice(), &mut dst).map(|()| (dst, src))
            })
            .map(|(a, _)| dst.write_all(a))??;

        Ok(())
    }
}

pub struct Asset<'a, R> {
    path: &'a str,
    data: R,
}

impl<'a, R> Asset<'a, R> {
    pub fn new(path: &'a str, data: R) -> Self {
        Asset { path, data }
    }
}

impl<'a> TryFrom<&'a str> for Asset<'a, BufReader<File>> {
    type Error = Error;

    fn try_from(value: &'a str) -> Result<Self, Self::Error> {
        File::open(value)
            .map(BufReader::new)
            .map(|data| Self::new(value, data))
            .map_err(Error::from)
    }
}

impl<'a, R: Read> Read for Asset<'a, R> {
    fn read(&mut self, buf: &mut [u8]) -> std::io::Result<usize> {
        self.data.read(buf)
    }
}

impl<'a, R: BufRead> BufRead for Asset<'a, R> {
    fn fill_buf(&mut self) -> std::io::Result<&[u8]> {
        self.data.fill_buf()
    }

    fn consume(&mut self, amt: usize) {
        self.data.consume(amt)
    }
}

pub struct Roxy;

impl Roxy {
    pub fn load(path: &str) -> Result<Asset<BufReader<File>>, Error> {
        path.try_into()
    }

    fn read_asset(
        asset: &mut Asset<BufReader<File>>,
        parser: &mut Parser,
    ) -> crate::result::Result<Vec<u8>> {
        let mut src = Vec::new();
        let mut dst = Vec::new();

        asset.data.read_to_end(&mut src)?;

        parser.parse(asset.path, &src, &mut dst)?;

        Ok(dst)
    }

    fn path_to_str<P: AsRef<Path>>(path: &P) -> std::io::Result<&str> {
        path.as_ref()
            .to_str()
            .ok_or_else(|| std::io::Error::new(std::io::ErrorKind::NotFound, "{path} is invalid"))
    }

    pub fn parse<P: AsRef<Path>>(path: &P, parser: &mut Parser) -> Result<Vec<u8>, Error> {
        let path = Self::path_to_str(path)?;
        Self::load(path).and_then(|mut a| Self::read_asset(&mut a, parser))
    }

    fn mkdir_and_open<P: AsRef<Path>>(path: &P) -> std::io::Result<File> {
        let path = path.as_ref();
        fs::create_dir_all(path.with_file_name(""))?;
        File::create(path)
    }

    fn to_formatted_path<P: AsRef<Path>>(path: &P) -> Result<PathBuf, Error> {
        let path = path.as_ref();
        match path.with_extension("").file_name() {
            Some(name) if name == "index" => Ok(path.to_path_buf()),
            Some(_) => Ok(path.with_extension("").join("index.html")),
            None => Err(anyhow::Error::msg("invalid path \"{path:?}\"").into()),
        }
    }

    pub fn process_file<P: AsRef<Path>>(
        input: &P,
        output: P,
        parser: &mut Parser,
    ) -> Result<(), Error> {
        let buf = Self::parse(input, parser)?;
        Self::mkdir_and_open(&Self::to_formatted_path(&output)?)
            .and_then(|mut f| f.write_all(&buf))
            .map_err(Error::from)
    }
}

#[cfg(test)]
mod tests {
    use super::Roxy;

    #[test]
    fn formatted_path() {
        let idx = Roxy::to_formatted_path(&"/test/index.html").unwrap();
        let name = Roxy::to_formatted_path(&"/i-love/cats.html").unwrap();

        assert_eq!(idx.to_str(), Some("/test/index.html"));
        assert_eq!(name.to_str(), Some("/i-love/cats/index.html"));
    }
}
